/*
 * (POP3Lite) SQLConf - 3lite POP3 Server (SQL configuration)
 * Copyright (C) 2000 The MadHouse Project
 *
 * This file is part of POP3Lite.
 *
 * POP3Lite is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * POP3Lite is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <pop3lite.h>
#include <glib.h>
#include <syslog.h>
#include <postgresql/libpq-fe.h>

#include "cfg.h"

static const char rcsid[]="$Id: sqlconf.c,v 1.3 2000/10/20 16:31:12 algernon Exp $";

int sqlconf_LTX_module_init ( P3LControl *control );
int sqlconf_LTX_module_done ( P3LControl *control );

static P3LHook_get_mailbox B_sqlconf_get_mailbox;

static char *sqlconf_get_mailbox ( P3LControl *control );
static int module_conf_init ( P3LControl *control );
static int module_mbox_init ( P3LControl *control );

/**
 * sqlconf_get_mailbox: get mailbox from an SQL table
 * @control: the usual control struct
 *
 * Determines a given user's mailbox from an SQL table
 *
 * Returns: the mailbox filename
 **/

static char *
sqlconf_get_mailbox ( P3LControl *control )
{
	char *user, *user_field, *box_field, *table, *tmp, *mb_name=NULL;
	PGconn *conn;
	PGresult *result;

#ifdef DEBUG
	control->system->log ( control, LOG_DEBUG, "%s:%d: get_mailbox (%s)",
			       __FILE__, __LINE__, P3L_GET_DATA ( "USER" ) );
#endif

	/*
	 * Get our variables ready
	 */
	user = P3L_GET_DATA ( "USER" );
	user_field = P3L_GET_FIRST_OPTION ( "SQLCONF.BOX.USER_FIELD" );
	box_field = P3L_GET_FIRST_OPTION ( "SQLCONF.BOX.BOX_FIELD" );
	table = P3L_GET_FIRST_OPTION ( "SQLCONF.BOX.TABLE" );
	conn = (PGconn *) g_hash_table_lookup ( control->data, "SQLCONF.BOX.CONNOBJ" );

	if ( user_field == NULL )
		user_field = "user";
	if ( box_field == NULL )
		box_field = "mailbox";

	/*
	 * If table is NULL, we're doomed
	 */
	if ( table == NULL )
	{
		if ( B_sqlconf_get_mailbox != NULL && 
		     p3l_is_enabled ( P3L_GET_FIRST_OPTION ( "SQLCONF.BOX.FALLBACK" ) ) )
			return (*B_sqlconf_get_mailbox)(control);
		else
			return NULL;
	}

	/*
	 * Execute the query:
	 * select [user_field], [mailbox_field] from [table] where
	 *	[user_field]='[user]';
	 */

	result = PQexec ( conn, g_strdup_printf ( "select %s, %s from %s where %s='%s';",
				user_field, box_field, table, user_field, user ) );

	/*
	 * Was it successful ?
	 */
	tmp = PQresultErrorMessage ( result );
	if ( tmp == NULL || (tmp && strlen ( tmp ) > 0 ) )
		control->system->log ( control, LOG_WARNING, "SQL message: %s", tmp );
	else
		mb_name = PQgetvalue ( result, 0, 1 );

	/*
	 * We don't need the result anymore
	 */

	PQclear(result);

	/*
	 * If the mailbox name is empty, we can fallback if allowed...
	 */

	if ( mb_name == NULL && B_sqlconf_get_mailbox != NULL && 
	     p3l_is_enabled ( P3L_GET_FIRST_OPTION ( "SQLCONF.BOX.FALLBACK" ) ) )
		return (*B_sqlconf_get_mailbox)(control);

	return g_strdup ( mb_name );
}

/**
 * module_conf_init: init SQLConf/conf
 * @control: the main control struct
 *
 * This loads the configuration from an SQL table.
 *
 * Returns: 0 on success, -1 otherwise
 **/

static int
module_conf_init ( P3LControl *control )
{
	PGconn *conn;
	PGresult *result;
	char *user, *db, *options, *table, *f_field, *f_value, *tmp;
	unsigned long i;

#ifdef DEBUG
	control->system->log ( control, LOG_DEBUG, "%s:%d: init mod-SQLConf/conf",
			       __FILE__, __LINE__ );
#endif

	/*
	 * Get ready...
	 */
	user = P3L_GET_FIRST_OPTION ("SQLCONF.CONF.USER");
	db = P3L_GET_FIRST_OPTION ("SQLCONF.CONF.DATABASE");
	options = P3L_GET_FIRST_OPTION ("SQLCONF.CONF.OTHER_OPTIONS");
	table = P3L_GET_FIRST_OPTION ("SQLCONF.CONF.TABLE");
	f_field = P3L_GET_FIRST_OPTION ("SQLCONF.CONF.FIELDS");
	f_value = P3L_GET_FIRST_OPTION ("SQLCONF.CONF.VALUES");

	if ( user == NULL )
		user = "postgres";
	if ( db == NULL )
		db = "template1";
	if ( options == NULL )
		options = "";
	if ( f_field == NULL )
		f_field = "field";
	if ( f_value == NULL )
		f_value = "value";
	if ( table == NULL )
		return -1;

	/*
	 * Connect to the database
	 */

	conn = PQconnectdb ( g_strdup_printf ( "user=%s dbname=%s %s", 
				user, db, options ) );

	/*
	 * Check for errors
	 */

	if ( PQstatus (conn) != CONNECTION_OK )
	{
		control->system->log ( control, LOG_WARNING,
				       "Connection error; SQL message: %s",
				       PQerrorMessage ( conn ) );
		PQfinish (conn);
		return -1;
	}

	/*
	 * Execute query:
	 *	select [f_field], [f_value] from table;
	 */

	result = PQexec ( conn, g_strdup_printf ("select %s, %s from %s;", 
					f_field, f_value, table ) );

	/*
	 * Was it succesful ?
	 */

	tmp = PQresultErrorMessage ( result );
	if ( tmp == NULL || (tmp && strlen ( tmp ) > 0 ) )
	{
		control->system->log ( control, LOG_WARNING,
				       "SELECT error; SQL message: %s", tmp );
		PQclear (result);
		PQfinish (conn);
		return -1;
	}

	/*
	 * Extract the things from the result
	 */

	for ( i = 0; i < PQntuples ( result ); i++ )
	{
		tmp = g_strdup ( PQgetvalue ( result, i, 0 ) );
		g_strup ( tmp );
		tmp = g_strdup_printf ( "%s = %s", tmp, PQgetvalue ( result, i, 1 ) );
		cfg_parse_line ( control, tmp, i, NULL, control->config );
		g_free ( tmp );
	}

	/*
	 * Cleanup
	 */

	PQclear(result);
	PQfinish(conn);

	return 0;
}

/**
 * module_mbox_init: init SQLConf/mbox
 * @conrol: the main control struct
 *
 * This initialises the SQL-based GET-MAILBOX hook.
 *
 * Returns: 0 on success, -1 otherwise
 **/

static int
module_mbox_init ( P3LControl *control )
{
	PGconn *conn;
	char *user, *db, *options;

#ifdef DEBUG
	control->system->log ( control, LOG_DEBUG, "%s:%d: init mod-SQLConf/mbox",
			       __FILE__, __LINE__ );
#endif

	/*
	 * Getting ready...
	 */

	user = P3L_GET_FIRST_OPTION ("SQLCONF.BOX.USER");
	db = P3L_GET_FIRST_OPTION ("SQLCONF.BOX.DATABASE");
	options = P3L_GET_FIRST_OPTION ("SQLCONF.BOX.OTHER_OPTIONS");

	if ( user == NULL )
		user = "postgres";
	if ( db == NULL )
		db = "template1";
	if ( options == NULL )
		options = "";

	/*
	 * Connect to the database
	 */

	conn = PQconnectdb ( g_strdup_printf ( "user=%s dbname=%s %s", 
				user, db, options ) );

	/*
	 * Are we there yet ?
	 */

	if ( PQstatus (conn) != CONNECTION_OK )
	{
		control->system->log ( control, LOG_WARNING,
				       "Connect error; SQL message: %s",
				       PQerrorMessage ( conn ) );
		PQfinish (conn);
		return -1;
	}

	/*
	 * Insert the connection object to the hash.
	 */

	g_hash_table_insert ( control->data, "SQLCONF.BOX.CONNOBJ", (gpointer) conn );

	/*
	 * Setup mailbox handler.
	 */

	B_sqlconf_get_mailbox = (P3LHook_get_mailbox)
		p3l_command_replace ( control->hooks, "GET-MAILBOX",
				      (gpointer) sqlconf_get_mailbox );

	return 0;
}

int
sqlconf_LTX_module_init ( P3LControl *control )
{
	char *tmp;
	int ret=-1;

#ifdef DEBUG
	control->system->log ( control, LOG_DEBUG, "%s:%d: init mod-SQLConf",
			       __FILE__, __LINE__ );
#endif

	tmp = P3L_GET_FIRST_OPTION ( "SQLCONF.BOX.ENABLE" );
	if ( tmp == NULL )
		tmp = "TRUE";

	if ( p3l_is_enabled ( tmp ) )
		module_conf_init ( control );

	tmp = P3L_GET_FIRST_OPTION ( "SQLCONF.CONF.ENABLE" );
	if ( tmp == NULL )
		tmp = "TRUE";

	if ( p3l_is_enabled ( tmp ) )
		ret = module_mbox_init ( control );

	return ret;
}

int
sqlconf_LTX_module_done ( P3LControl *control )
{
	PGconn *conn;

#ifdef DEBUG
	control->system->log ( control, LOG_DEBUG, "%s:%d: done mod-SQLConf",
			       __FILE__, __LINE__ );
#endif

	/*
	 * Destroy the database connection.
	 */
	conn = (PGconn *) g_hash_table_lookup ( control->data, "SQLCONF.BOX.CONNOBJ" );
	if ( conn != NULL )
	{
		PQfinish ( conn );

		g_hash_table_insert ( control->hooks, "GET-MAILBOX",
				      (gpointer) B_sqlconf_get_mailbox );
	}

	return 0;
}
